/*
メッシュ構成のp2pクライアント
*/
/*
  基本的な使い方
  P2P_Clientをnewする。
  後はコールバックが呼ばれる
  url,callbacks,channels,rtc_config=null
  url  シグナリングサーバ
  callbacks
    onStateChange state
    onJoin id connection
    onLeave id connection
    onDataOpen id usage
    onData     id usage data  -> channelsに移行
    onDataClose id usage
  channels 名前をキーとする以下の要素の連想配列 
    type chat or  file  or transform or rpc
    onDataOpen() id 
    onData() id  data
    onDataClose() id
  rtc_config peerconnectionのconfig
*/


export class P2P_Connection{
  constructor(sender,target,client){
    this.sender = sender;
    this.target = target;
    this.client = client;
    this.state = "begin";
    this.dic_channel = {};//usage to channel
    client.dic_connection[target]=this;
    this.createPeerConnection();

  };
  
  send(label,data){
    if(label in this.dic_channel){
      this.dic_channel[label].send(data);
    }else{
      console.log("channel "+ label + " not open");
    }
  };

  createPeerConnection() {
    let that=this;
    let sender = that.sender;
    let target = that.target;
    let client = that.client;
    this.pc = new RTCPeerConnection(client.rtc_config);

    // ICE candidate 取得時のイベントハンドラを登録
    that.pc.onicecandidate = function(evt) {
      if ( evt.candidate ) {
        console.log(evt.candidate);
        let message = {
          mode: "candidate"
          , Candidate: evt.candidate.candidate
          , sdpMLineIndex:evt.candidate.sdpMLineIndex
          , sdpMid:evt.candidate.sdpMid
          ,sender:sender
          ,target:target
        };
        that.client.socket.send(JSON.stringify(message));
        that.state = 'Collecting ICE candidates';
      } else {
      }
    };

    that.pc.onconnectionstatechange = function(evt) {
      switch(that.pc.connectionState) {
      case "connected":
        that.state = 'connected';
        that.client._onJoin(target,that);
        break;
      case "disconnected":
      case "failed":
        that.state = 'disconnected';
        that.client._onLeave(target,that);
        delete that.client.dic_connection[target];
        break;
      case "closed":
        that.state = 'closed';
        delete that.client.dic_connection[target];
        break;
      }
    };
    
    that.pc.ondatachannel = function(evt) {
      console.log('Data channel created:', evt);
      that.setupDataChannel(evt.channel,that.client.channels);
      //let dataChannel = evt.channel;
      //that.dc=evt.channel;
    };
    that.pc.onicegatheringstatechange = function()
    {
      console.log( "Event : ICE gathering state change" );
      console.log( "- ICE gathering state : ", that.pc.iceGatheringState );

      if( "complete" === that.pc.iceGatheringState )
      {
        let localSDP = that.pc.localDescription.sdp;
        if( that.pc.localDescription.type=="offer"){
        }else if(that.pc.localDescription.type=="answer"){
        }
      }
    };
    return that.pc;
  }
  setupDataChannel(dc,channels) {
    let that=this;
    let dcinfo = channels[dc.label];
    
    dc.onerror = function(error) {
      console.log('Data channel error:', error);
    };
    dc.onmessage = function(evt) {
      console.log('Data channel message:', evt.data);
      let msg = evt.data;
      if("onData" in dcinfo){
        dcinfo.onData(that.target,msg);
      }
    };
    dc.onopen = function(evt) {
      that.dic_channel[dc.label]=dc;
      if("onDataOpen" in dcinfo){
        dcinfo.onDataOpen(that.target);
      }
      console.log('Data channel opened:', evt);
    };
    dc.onclose = function(evt) {
      if("onDataClose" in dcinfo){
        dcinfo.onDataClose(that.target);
      }
      console.log('Data channel closed.');
    };
  }
} 

/*
const dataChannelOptions = {
  ordered: false,
}
*/




export const pack_transform=(src)=>{
  var buffer = new ArrayBuffer(3*4+3*4+8);
  var view = new DataView(buffer);
  view.setFloat32(0,src.position.x);
  view.setFloat32(4,src.position.y);
  view.setFloat32(8,src.position.z);
  view.setFloat32(12,src.rotation.x);
  view.setFloat32(16,src.rotation.y);
  view.setFloat32(20,src.rotation.z);
  view.setFloat64(24,src.ts);
  return buffer;
}

export const unpack_transform = (buffer)=>{
  var view = new DataView(buffer);
  var res = {
    position:{x:0,y:0,z:0},
    rotation:{x:0,y:0,z:0},
    ts:0
  };
  res.position.x = view.getFloat32(0);
  res.position.y = view.getFloat32(4);
  res.position.z = view.getFloat32(8);
  res.rotation.x = view.getFloat32(12);
  res.rotation.y = view.getFloat32(16);
  res.rotation.z = view.getFloat32(20);
  res.ts = view.getFloat64(24);
  return res;
}

export const pack_position = (src)=>{
  var buffer = new ArrayBuffer(3*4);
  var view = new DataView(buffer);
  view.setFloat32(0,src.x);
  view.setFloat32(4,src.y);
  view.setFloat32(8,src.z);
  return buffer;
}

export const unpack_position=(buffer)=>{
  var view = new DataView(buffer);
  var res = {
    x:0,y:0,z:0
  };
  res.x = view.getFloat32(0);
  res.y = view.getFloat32(4);
  res.z = view.getFloat32(8);
  return res;
}



const default_rtc_config ={'iceServers': [
  { urls: "stun:stun.l.google.com:19302" },
  { urls: "stun:stun1.l.google.com:19302" },
]};

export class P2P_Client{
  constructor(url,callbacks,channels,rtc_config=null){
    this.url = url;
    this.my_id = "";
    this.state = "begin";
    this.dic_connection = {};//id to P2P_Connection
    this.socketReady = false;
    this.channels = channels;
    this.room = location.protocol+"/"+location.host+location.pathname;
    this.rtc_config = rtc_config?rtc_config:default_rtc_config;
    if("onStateChange" in callbacks){
      this.onStateChange = callbacks.onStateChange;
    };
    if("onJoin" in callbacks){
      this.onJoin = callbacks.onJoin;
    };
    if("onLeave" in callbacks){
      this.onLeave = callbacks.onLeave;
    };

    this.init();
  }
  
  init(){
    let that = this;
    this.socket =  new WebSocket(this.url);
    this.socket.addEventListener('open',(e)=>{
      console.log('Socket 接続成功');
      that.socketReady = true;
      let message= {
          mode:"join",
          room:that.room
      };
      that.socket.send(JSON.stringify(message));
    });
    that.socket.addEventListener('message',(e)=>{
      that.onMessage(JSON.parse(e.data));
      console.log(e.data);
    });
  }
  send(id,label,data){
    //特定の相手に送信
    if(id in this.dic_connection){
      this.dic_connection[id].send(label,data);
    }else{
      console.log("no user ",id);
    }
  };

  broadcast(label,data){
    //全体に送信
    for(let id in this.dic_connection){
      this.dic_connection[id].send(label,data);
    }
  };
  
  onMessage(evt) {
    let that = this;
    console.log(evt);
    if (evt.mode === 'welcome') {
      that.my_id=evt.you;
      //先輩方に挨拶
      for(let target of evt.members){
        that.offer(that.my_id,target);
      }
    }else if (evt.mode === 'offer') {
        that.onOffer(evt);   
    } else if (evt.mode === 'answer') {
        that.onAnswer(evt);
    } else if (evt.mode === 'candidate' ) {
        that.onCandidate(evt);
    }
  }
  
  _onJoin(target,connection){
    if(this.onJoin){
      this.onJoin(target,connection);
    }
  }
  _onLeave(target,connection){
    if(this.onLeave){
      this.onLeave(target,connection);
    }
  }
  
  _onStateChange(st){
    this.state = st;
    if(this.onStateChange){
      this.onStateChange(st);
    }
  }
  
  
  offer(sender,target){
    let that = this;
    console.log("offer",sender,target);
    // 新しい RTCPeerConnection を作成する
    let connection = new P2P_Connection(sender,target,this);
    let peerConnection = connection.pc;
    
    // Data channel を生成
    for(let label in that.channels){
      let options = {
      };
      let dcinfo = that.channels[label];

      if(dcinfo.type=="chat" || dcinfo.type=="file"){
        //チャットやファイル転送の場合はデフォルト
        options.ordered = true;
      }else{
        //位置同期やRPCの場合は速度優先
        options.ordered = false;
      }
      let dataChannel = peerConnection.createDataChannel(label, options);
      connection.setupDataChannel(dataChannel,that.channels);
    }

    // Offer を生成する
    peerConnection.createOffer().then(function(sessionDescription) {
      console.log('createOffer() succeeded.');
      let localSDP = sessionDescription.sdp;
      return peerConnection.setLocalDescription(sessionDescription);
    }).then(function() {
      // setLocalDescription() が成功した場合
      // Trickle ICE ではここで SDP を相手に通知する
      // Vanilla ICE では ICE candidate が揃うのを待つ
      //シグナリングはpeerConnectionで処理している
      let message= {
          mode:"offer"
          ,sender:sender
          ,target:target
          ,sdp:peerConnection.localDescription
      };
      that.socket.send(JSON.stringify(message));
      console.log('setLocalDescription() succeeded.');
    }).catch(function(err) {
      console.error('setLocalDescription() failed.', err);
    });
    that._onStateChange("offer created");
  }

  onOffer(evt) {
    let remoteSDP = evt.sdp;
    console.log("onOffer",evt);
    let that = this;
    let sender = evt.sender;
    let target = evt.target;

    // Peer Connection が未生成の場合，SDP を Offer と見なす
    let offer = new RTCSessionDescription({
      type: 'offer',
      sdp: remoteSDP.sdp,
    });
    // Peer Connection を生成
    let connection = new P2P_Connection(that.my_id,evt.sender,this);
    let peerConnection = connection.pc;
    
    //trickleの場合この時点ではまだ設定できない？
    peerConnection.setRemoteDescription(offer).then(function() {
      console.log('setRemoteDescription() succeeded.');
      return peerConnection.createAnswer();
    }).then(function(sessionDescription ) {
      return peerConnection.setLocalDescription( sessionDescription );
    }).then(function() {
      that._onStateChange('answer created');
      let message= {
        mode:"answer"
        ,sender:evt.target
        ,target:evt.sender
        ,sdp:peerConnection.localDescription
      };
      that.socket.send(JSON.stringify(message));
    }).catch(function(err) {
      console.error('setRemoteDescription() failed.', err);
    });


  }

  onAnswer(evt) {
    let that = this;
    let remoteSDP = evt.sdp;
    let connection = that.dic_connection[evt.sender];
    let peerConnection = connection.pc;
    // Peer Connection が生成済みの場合，SDP を Answer と見なす
    let answer = new RTCSessionDescription({
      type: 'answer',
      sdp: remoteSDP.sdp,
    });
    peerConnection.setRemoteDescription(answer).then(function() {
      console.log('setRemoteDescription() succeeded.');
    }).catch(function(err) {
      console.error('setRemoteDescription() failed.', err);
    });
  }
  
  onCandidate(evt) {
    let that = this;
    if(! ('sdpMLineIndex' in evt)){
      console.log("ignore candidate by null sdpMLineIndex");
      return;
    }
    let connection =that.dic_connection[evt.sender];
    let peerConnection = connection.pc;
    var candidate = new RTCIceCandidate({sdpMLineIndex:evt.sdpMLineIndex, sdpMid:evt.sdpMid, candidate:evt.Candidate});
    peerConnection.addIceCandidate(candidate);
  }
  
  
}

